<?php

namespace Datamints\DatamintsPiwik\Service;

/***************************************************************
 *  Copyright notice
 *
 *  (c) 2015 Eva Rieger <e.rieger@datamints.com>, datamints GmbH
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Extbase\Utility\LocalizationUtility;

/**
 * Service class for exporting data from sys_log and sys_note to Piwik as annotations.
 */
class PiwikExportService extends PiwikConnectService {

	const EXPORT_TYPE_NOTEXPORTED = 0;

	const EXPORT_TYPE_SUCCESS = 1;

	const EXPORT_TYPE_ERROR = 2;

	/**
	 * Repository class for sys_note table
	 *
	 * @var \TYPO3\CMS\SysNote\Domain\Repository\SysNoteRepository
	 * @inject
	 */
	protected $sysNoteRepository;

	/**
	 * Sets the 'processed'-timestamp for notes and logs to current timestamp. This ensures that old logs will be not exported the first time.
	 */
	public function invalidate() {
		$this->initialize();
		$this->db->exec_UPDATEQuery('sys_note', '', array('tx_datamintspiwik_processed' => time()));
		$this->db->exec_UPDATEQuery('sys_log', '', array('tx_datamintspiwik_processed' => time()));
	}

	/**
	 * Export relevant data to Piwik.
	 *
	 * @param boolean $isSysNotesExport
	 */
	public function export($isSysNotesExport) {
		$this->initialize();

		if (!intval($this->settings['enable'])) {
			return;
		}

		$messages = array();
		if ($isSysNotesExport) {
			$messages = array_merge($messages, $this->getNotes($this->settings['export']['sys_notes']));
		}
		else {
			$messages = array_merge($messages, $this->getLogs($this->settings['export']['sys_log']));
		}

		foreach ($messages as $message) {
			if ($this->sendMessage($message)) {
				$this->updateMessage($message, self::EXPORT_TYPE_SUCCESS);
			}
			else {
				$this->updateMessage($message, self::EXPORT_TYPE_ERROR);
			}
		}
	}

	/**
	 * Returns messages from sys_note objects to be exported.
	 *
	 * @param array $config
	 * @return array
	 */
	protected function getNotes($config) {
		$additionalWhere = '';
		if (isset($config['type']) && trim($config['type']) !== '') {
			$additionalWhere = ' AND category in (' . $this->prepareForInStatement($config['type'], 'sys_note') . ')';
		}
		$statement = 'SELECT * FROM sys_note WHERE deleted = 0 AND (tx_datamintspiwik_processed = 0 OR tx_datamintspiwik_exported = 2) AND personal = 0' . $additionalWhere;

		$query = $this->sysNoteRepository->createQuery();
		$query->statement($statement);
		$res = $query->execute();
		$messages = array();
		foreach ($res as $note) {
			// Only handle notes from current microsite.
			// The event type is always 1 because it is an interal note.
			if (!$this->settings['rootpid'] || $this->checkPid($note->getPid(), $this->settings['rootpid'])) {
				$messages[] = array(
					'message' => $note->getSubject() . ': ' . $note->getMessage(),
					'date' => $note->getModificationDate()->format('Y-m-d'),
					'table' => 'sys_note',
					'uids' => array($note->getUid()),
					'type' => 1,
				);
			}
		}

		return $messages;
	}

	/**
	 * Returns entries from sys_log, grouped by affected table, uid and day.
	 *
	 * @param array $config
	 * @return array
	 */
	protected function getLogs($config) {
		$additionalWhere = '';
		if (isset($config['type']) && trim($config['type']) !== '') {
			$additionalWhere = ' AND type IN (' . $this->prepareForInStatement($config['type'], 'sys_log') . ')';
		}

		$res = $this->db->exec_SELECTQuery('sys_log.*,	sys_history.fieldlist, sys_history.history_data', 'sys_log LEFT JOIN sys_history ON sys_log.uid = sys_history.sys_log_uid', '(sys_log.tx_datamintspiwik_processed = 0 OR sys_log.tx_datamintspiwik_exported = 2)' . $additionalWhere);
		$messages = array();
		while ($row = $this->db->sql_fetch_assoc($res)) {
			// Only handle pages in the current microsite.
			if ($this->settings['rootpid'] && !$this->checkPid($row['event_pid'], $this->settings['rootpid'])) {
				continue;
			}

			// Merge logs by day.
			$date = date('Y-m-d', $row['tstamp']);
			if ($messages[$row['tablename']][$row['recuid']][$date]) {
				$messages[$row['tablename']][$row['recuid']][$date] = $this->mergeLogs($messages[$row['tablename']][$row['recuid']][$date], $row);
			}
			else {
				$row['uids'] = array($row['uid']);
				$row['actions'] = array($row['action']);
				$row['detailsArr'] = array($row['details']);
				$messages[$row['tablename']][$row['recuid']][$date] = $row;
			}

		}

		$outMessages = array();
		foreach ($messages as $table => $tableData) {
			foreach ($tableData as $uid => $uidData) {
				foreach ($uidData as $date => $data) {
					if ($message = $this->canSendLogMessage($config, $data)) {
						$outMessages[] = array(
							'message' => $message['note'],
							'date' => $date,
							'table' => 'sys_log',
							'uids' => $data['uids'],
							'type' => $message['type'],
						);
					}
					else {
						$this->updateMessage(array('table' => 'sys_log', 'uids' => $data['uids']), self::EXPORT_TYPE_NOTEXPORTED);
					}
				}
			}
		}

		return $outMessages;
	}

	/**
	 * Merges two log entries:
	 * data is put in arrays, these arrays are serialized and merged, then written back.
	 *
	 * @param array $old
	 * @param array $new
	 * @return array
	 */
	protected function mergeLogs($old, $new) {
		$old['uids'][] = $new['uid'];
		$old['actions'][] = $new['action'];
		$old['detailsArr'][] = $new['details'];
		$old['log_data'] = $new['log_data'];
		$old['fieldlist'] = implode(',', array_unique(array_merge(explode(',', $old['fieldlist']), array_merge(explode(',', $new['fieldlist'])))));

		if (!$old['history_data']) {
			$old['history_data'] = serialize(array('oldRecord' => array(), 'newRecord' => array()));
		}
		if (!$new['history_data']) {
			$new['history_data'] = serialize(array('oldRecord' => array(), 'newRecord' => array()));
		}
		$oldData = unserialize($old['history_data']);
		$newData = unserialize($new['history_data']);

		// For source data, $oldData is the master, for output data, $newData is the master;
		// this way you get data from oldest to newest entry.
		$oldData['oldRecord'] = array_merge($newData['oldRecord'], $oldData['oldRecord']);
		$oldData['newRecord'] = array_merge($oldData['oldRecord'], $newData['oldRecord']);
		$old['history_data'] = serialize($oldData);

		return $old;
	}

	/**
	 * Check if given data can be sent to Piwik.
	 *
	 * If the message is worth sending, the method returns a consolidated version of it.
	 *
	 * @param array $config
	 * @param array $data
	 * @return boolean|array
	 */
	protected function canSendLogMessage($config, $data) {
		if (!in_array($data['tablename'], explode(',', $config['tables']))) {
			return FALSE;
		}

		$moved = FALSE;
		$use = 0;

		// Check if record was created and immediately deleted.
		if ($data['type'] == '1') {
			foreach ($data['actions'] as $action) {
				switch ($action) {
					case 1:
						$use = 1 << 10;
						break;

					case 2:
						$use++;
						break;

					case 4:
						$use++;
						$moved = TRUE;
						break;

					case 3:
						if ($use) {
							$use = 0;
						}
						else {
							$use = -1;
						}
						break;
				}
			}
		}

		// Calculate the final action, depending on $use, so that we only have on final action to report:
		// - multiple edits => single edit
		// - create and edit => create
		// - create and delete => nothing
		$note = '';
		$logData = unserialize($data['log_data']);
		if ($use >= 1 << 10) {
			$action = 1;
			$note = "'" . $logData[0] . "'" . ' (' . $data['tablename'] . '/' . $data['recuid'] . ')';
		}
		else {
			if ($use < 0) {
				$action = 3;
				$note = "'" . $logData[0] . "'" . ' (' . $data['tablename'] . '/' . $data['recuid'] . ')';
			}
			else {
				if ($use == 0) {
					$action = 0;
				}
				else {
					$note = "'" . $logData[0] . "'" . ' (' . $data['tablename'] . '/' . $data['recuid'] . '): ';
					$action = 2;
				}
			}
		}

		// @todo Handle record moves ($moved = true)

		// Filter out irrelevant actions.
		if (!GeneralUtility::inList($config[$data['tablename']]['actions'], $action)) {
			return FALSE;
		}

		// Update is only important for certain fields.
		$updateMessages = array();
		if ($action == 2) {
			$updateFields = array_intersect(explode(',', $config[$data['tablename']]['updateFields']), explode(',', $data['fieldlist']));
			if (!count($updateFields)) {
				return FALSE;
			}

			$historyData = unserialize($data['history_data']);
			foreach ($updateFields as $field) {
				$updateMessages[] = $field . ': ' . $historyData['oldRecord'][$field] . ' => ' . $historyData['newRecord'][$field];
			}

			$note .= implode(', ', $updateMessages);
		}
		$message = array(
			'note' => $note,
			'type' => $action,
		);

		return $message;
	}

	/**
	 * Creates a comma-separated string of SQL-escaped IN-Statement
	 *
	 * @param string $string SQL IN-statement
	 * @param string $table Table name for which to quote
	 * @return string
	 */
	protected function prepareForInStatement($string, $table) {
		return implode(',', $this->db->fullQuoteArray(explode(',', $string), $table));
	}

	/**
	 * Send a message to Piwik.
	 *
	 * @param array $message
	 * @return boolean
	 */
	protected function sendMessage($message) {
		$result = $this->piwikRepository->addEvent($message['date'], $message['message'], $message['type']);

		if (!isset($result) || $result === FALSE) {
			return FALSE;
		}

		if (array_key_exists('result', $result) && $result['result'] == 'error') {
			throw new \Exception(LocalizationUtility::translate('piwik.export.missing_datamints', 'datamints_piwik'));
		}

		return TRUE;
	}

	/**
	 * Mark an entry as exported.
	 *
	 * @param array $note
	 * @param int $code
	 */
	protected function updateMessage($note, $code) {
		if ($note['table'] && is_array($note['uids']) && count($note['uids'])) {
			$sql = 'UPDATE ' . $note['table'] . ' SET tx_datamintspiwik_exported = ' . $this->db->quoteStr($code, $note['table']) . ', tx_datamintspiwik_processed =   ' . time() . ' WHERE  uid in (' . implode(',', $note['uids']) . ')';
			$this->db->sql_query($sql);
		}
	}

	/**
	 * Check if $uid is in the page tree starting at $rootPid, limited by depth $depth.
	 *
	 * @todo Use TYPO3 model classes for data retrieval?
	 *
	 * @param string $uid
	 * @param string $rootPid
	 * @param int $depth
	 * @return boolean
	 */
	protected function checkPid($uid, $rootPid, $depth = 0) {
		if ($depth > 100) {
			return FALSE;
		}

		if ($uid == 0) {
			return FALSE;
		}

		if ($uid == $rootPid) {
			return TRUE;
		}

		$row = $this->db->exec_SELECTgetSingleRow('pid', 'pages', 'uid=' . intval($uid));

		return $this->checkPid($row['pid'], $rootPid, $depth + 1);
	}

}
